/*
 * Copyright (c) 2019 gematik - Gesellschaft für Telematikanwendungen der Gesundheitskarte mbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.gematik.ti.vauchannel.protocol.helpers;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import de.gematik.rs.vau.*;
import de.gematik.ti.vauchannel.protocol.VAUProtocolCrypto;
import de.gematik.ti.vauchannel.protocol.VAUProtocolSession;
import org.bouncycastle.jce.provider.BouncyCastleProvider;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.nio.charset.StandardCharsets;
import java.security.PublicKey;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;


public class VAUProtocolHelpers {

    static final Logger logger = LoggerFactory.getLogger(VAUProtocolHelpers.class);
    static ObjectMapper mapper = new ObjectMapper();

    public static List<CipherConfiguration> getCipherConfiguration() {
        List<CipherConfiguration> cipherConfiguration = new ArrayList<>();
        cipherConfiguration.add(CipherConfiguration.AES_256_GCM_BRAINPOOL_P_256_R_1_SHA_256);
        return cipherConfiguration;
    }

    public static List<CipherConfiguration_> getCipherConfiguration_() {
        List<CipherConfiguration_> cipherConfiguration = new ArrayList<>();
        cipherConfiguration.add(CipherConfiguration_.AES_256_GCM_BRAINPOOL_P_256_R_1_SHA_256);
        return cipherConfiguration;
    }

    public static void checkClientSignature(VAUClientSigFin vAUClientSigFin, VAUProtocolCrypto crypto) {
        boolean signatureOk;
        try {
            logger.info("ClientSignature: " + vAUClientSigFin.getSignature());

            byte[] signature = Base64.decode(vAUClientSigFin.getSignature());

            byte[] message = concat(vAUClientSigFin.getVAUClientHelloDataHash().getBytes(StandardCharsets.UTF_8), vAUClientSigFin.getVAUServerHelloDataHash().getBytes(StandardCharsets.UTF_8));

            byte[] certBytes = Base64.decode(vAUClientSigFin.getCertificate());
            CertificateFactory certFactory = CertificateFactory.getInstance("X.509", BouncyCastleProvider.PROVIDER_NAME);

            InputStream in = new ByteArrayInputStream(certBytes);
            X509Certificate cert = (X509Certificate) certFactory.generateCertificate(in);
            PublicKey pk = cert.getPublicKey();

            signatureOk = crypto.verify(message, signature, pk);

        } catch (Exception e) {
            logger.error(e.getMessage(), e);
            throw new RuntimeException("Exception");
        }


        if (!signatureOk) {
            // Falls die Signaturprüfung ein nicht-positives Prüferergebnis liefert, so MUSS der Server
            // mit einer VAUServerError-Nachricht mit der Fehlermeldung "Signature from VAUClientSigFin invalid"
            // antworten und die weitere Protokolldurchführung abbrechen.
            throw new RuntimeException("Signature from VAUClientSigFin invalid");
        }
    }

    public static void checkServerSignature(VAUServerHello vAUServerHello, VAUProtocolCrypto crypto) {
        boolean signatureOk;
        try {
            logger.info("ServerSignature: " + vAUServerHello.getSignature());

            byte[] signature = Base64.decode(vAUServerHello.getSignature());


            byte[] message = vAUServerHello.getData().getBytes(StandardCharsets.UTF_8);

            byte[] certBytes = Base64.decode(vAUServerHello.getCertificate());
            CertificateFactory certFactory = CertificateFactory.getInstance("X.509", BouncyCastleProvider.PROVIDER_NAME);

            InputStream in = new ByteArrayInputStream(certBytes);
            X509Certificate cert = (X509Certificate) certFactory.generateCertificate(in);
            PublicKey pk = cert.getPublicKey();

            signatureOk = crypto.verify(message, signature, pk);

        } catch (Exception e) {
            throw new RuntimeException("Exception");
        }


        if (!signatureOk) {
            throw new RuntimeException("Signature from VAUServerHello invalid");
        }
    }

    public static void checkFinishedDataFromClient(VAUProtocolSession serverSession, VAUClientSigFin vAUClientSigFin, byte[] vauServerHelloDataHash) {
        // Falls die Prüfung des "FinishedData"-Feld ein nicht-positives Prüferergebnis liefert,
        // so MUSS der Server mit einer VAUServerError-Nachricht mit der Fehlermeldung
        // "VAUClientSigFin invalid" antworten und die weitere Protokolldurchführung abbrechen.
        String errorMessage = "VAUClientSigFin invalid";
        byte[] finishedDataBytes = Base64.decode(vAUClientSigFin.getFinishedData());
        byte[] keyidOnServer = serverSession.getKeyID();
        byte[] keyidFromClient = new byte[keyidOnServer.length];
        System.arraycopy(finishedDataBytes, 0, keyidFromClient, 0, keyidOnServer.length);
        if (!Arrays.equals(keyidOnServer, keyidFromClient)) {
            logger.error("KeyID on Server " + Base64.encode2String(keyidOnServer) + " not equals KeyID on Client " + Base64.encode2String(keyidFromClient));
            throw new RuntimeException(errorMessage);
        }
        byte[] encryptedMessageFromClient = new byte[finishedDataBytes.length - keyidOnServer.length];
        System.arraycopy(finishedDataBytes, keyidOnServer.length, encryptedMessageFromClient, 0, encryptedMessageFromClient.length);
        byte[] decryptedMessage = null;
        try {
            decryptedMessage = AESGCM.decrypt(encryptedMessageFromClient, serverSession.getSymKey());
        } catch (Exception e) {
            logger.error(e.getMessage(), e);
            throw new RuntimeException(errorMessage);
        }
        // "VAUClientSigFin"
        // || unkodierter Hashwert aus "VAUClientHelloDataHash"
        // || unkodierter Hashwert aus "VAUServerHelloDataHash"
        byte[] concatenatedHashesUncoded = VAUProtocolHelpers.concat(
                Base64.decode(vAUClientSigFin.getVAUClientHelloDataHash()),
                Base64.decode(vAUClientSigFin.getVAUServerHelloDataHash()));
        byte[] expectedMessage = VAUProtocolHelpers.concat("VAUClientSigFin".getBytes(StandardCharsets.UTF_8), concatenatedHashesUncoded);
        if (expectedMessage.length != 79) {
            logger.error("message.length = " + expectedMessage.length + ", expected is 79");
            throw new RuntimeException(errorMessage);
        }
        if (!Arrays.equals(decryptedMessage, expectedMessage)) {
            logger.error("decryptedMessage = " + decryptedMessage + " not equals expectedMessage = " + expectedMessage);
            throw new RuntimeException(errorMessage);
        }
    }

    public static void checkFinishedDataFromServer(VAUProtocolSession session, VAUServerFin vAUServerFin) {
        // Der Client MUSS beim Empfang der VAUServerFin-Nachricht prüfen, ob der Wert im "FinishedData"-Feld
        // der nach A_16899 zu erwartenden Wert entspricht.
        // Falls nein, so MUSS der Client den weiteren Protokollablauf abbrechen (vgl. A_16849).

        byte[] vauClientHelloDataHash = session.getClientHelloDataHash();
        byte[] vauServerHelloDataHash = session.getServerHelloDataHash();


        String errorMessage = "VAUServerFin invalid";
        byte[] finishedDataBytes = Base64.decode(vAUServerFin.getFinishedData());
        byte[] keyidOnServer = session.getKeyID();
        byte[] keyidFromServer = new byte[keyidOnServer.length];
        System.arraycopy(finishedDataBytes, 0, keyidFromServer, 0, keyidOnServer.length);
        if (!Arrays.equals(keyidOnServer, keyidFromServer)) {
            logger.error("KeyID on Server " + Base64.encode2String(keyidOnServer) + " not equals KeyID on Client " + Base64.encode2String(keyidFromServer));
            throw new RuntimeException(errorMessage);
        }
        byte[] encryptedMessageFromServer = new byte[finishedDataBytes.length - keyidOnServer.length];
        System.arraycopy(finishedDataBytes, keyidOnServer.length, encryptedMessageFromServer, 0, encryptedMessageFromServer.length);
        byte[] decryptedMessage = null;
        try {
            decryptedMessage = AESGCM.decrypt(encryptedMessageFromServer, session.getSymKey());
        } catch (Exception e) {
            logger.error(e.getMessage(), e);
            throw new RuntimeException(errorMessage);
        }
        // "VAUServerFin"
        // || unkodierter Hashwert aus "VAUClientHelloDataHash"
        // || unkodierter Hashwert aus "VAUServerHelloDataHash"
        // Diese Zeichenkette MUSS 12+32+32=76 Bytes lang sein.
        byte[] concatenatedHashesUncoded = VAUProtocolHelpers.concat(
                vauClientHelloDataHash,
                vauServerHelloDataHash);
        byte[] expectedMessage = VAUProtocolHelpers.concat("VAUServerFin".getBytes(StandardCharsets.UTF_8), concatenatedHashesUncoded);
        if (expectedMessage.length != 76) {
            logger.error("message.length = " + expectedMessage.length + ", expected is 76");
            throw new RuntimeException(errorMessage);
        }
        if (!Arrays.equals(decryptedMessage, expectedMessage)) {
            logger.error("decryptedMessage = " + decryptedMessage + " not equals expectedMessage = " + expectedMessage);
            throw new RuntimeException(errorMessage);
        }
    }

    public static String prettyJson(String str) {
        try {
            Object json = mapper.readValue(str, Object.class);
            return mapper.writerWithDefaultPrettyPrinter().writeValueAsString(json);
        } catch (Exception ex) {
            logger.error(Arrays.toString(ex.getStackTrace()));
        }
        return str;
    }

    public static void logJSON(Object o) {
        String str = null;
        if (o instanceof String) {
            str = (String) o;
        } else {
            try {
                str = mapper.writeValueAsString(o);
            } catch (JsonProcessingException ex) {
                logger.error("", ex);
            }
        }
        logger.info(prettyJson(str));
    }

    public static byte[] concat(byte[] a, byte[] b) {
        byte[] c = new byte[a.length + b.length];
        System.arraycopy(a, 0, c, 0, a.length);
        System.arraycopy(b, 0, c, a.length, b.length);
        return c;
    }
}
